@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # Copyright (c) the purl authors
+# SPDX-License-Identifier: MIT
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -23,38 +24,15 @@
 # Visit https://github.com/package-url/packageurl-python for support and
 # download.
 
-
-from __future__ import absolute_import
-from __future__ import print_function
-from __future__ import unicode_literals
-
-from collections import namedtuple
-from collections import OrderedDict
 import string
+from collections import namedtuple
+from urllib.parse import quote as _percent_quote
+from urllib.parse import unquote as _percent_unquote
+from urllib.parse import urlsplit as _urlsplit
 
-# Python 2 and 3 support
-try:
-    # Python 2
-    from urlparse import urlsplit as _urlsplit
-    from urllib import quote as _percent_quote
-    from urllib import unquote as _percent_unquote
-except ImportError:
-    # Python 3
-    from urllib.parse import urlsplit as _urlsplit
-    from urllib.parse import quote as _percent_quote
-    from urllib.parse import unquote as _percent_unquote
-
-# Python 2 and 3 support
-try:
-    # Python 2
-    unicode
-    basestring = basestring  # NOQA
-    bytes = str  # NOQA
-    str = unicode  # NOQA
-except NameError:
-    # Python 3
-    unicode = str  # NOQA
-    basestring = (bytes, str,)  # NOQA
+# Python 3
+unicode = str  # NOQA
+basestring = (bytes, str,)  # NOQA
 
 """
 A purl (aka. Package URL) implementation as specified at:
@@ -159,7 +137,7 @@ def normalize_qualifiers(qualifiers, encode=True):  # NOQA
     Raise ValueError on errors.
     """
     if not qualifiers:
-        return None if encode else OrderedDict()
+        return None if encode else dict()
 
     if isinstance(qualifiers, basestring):
         if not isinstance(qualifiers, unicode):
@@ -206,7 +184,7 @@ def normalize_qualifiers(qualifiers, encode=True):  # NOQA
                 "A qualifier key cannot start with a number: {}".format(repr(key)))
 
     qualifiers = sorted(qualifiers.items())
-    qualifiers = OrderedDict(qualifiers)
+    qualifiers = dict(qualifiers)
     if encode:
         qualifiers = ['{}={}'.format(k, v) for k, v in qualifiers.items()]
         qualifiers = '&'.join(qualifiers)
@@ -260,8 +238,14 @@ class PackageURL(namedtuple('PackageURL', _components)):
             raise ValueError('Invalid purl: {} is a required argument.'
                              .format(key))
 
-        strings = dict(type=type, namespace=namespace, name=name,
-                       version=version, subpath=subpath)
+        strings = dict(
+            type=type,
+            namespace=namespace,
+            name=name,
+            version=version,
+            subpath=subpath,
+        )
+
         for key, value in strings.items():
             if value and isinstance(value, basestring) or not value:
                 continue
@@ -275,9 +259,15 @@ class PackageURL(namedtuple('PackageURL', _components)):
         type, namespace, name, version, qualifiers, subpath = normalize(# NOQA
             type, namespace, name, version, qualifiers, subpath, encode=None)
 
-        return super(PackageURL, self).__new__(PackageURL, type=type,
-            namespace=namespace, name=name, version=version,
-            qualifiers=qualifiers, subpath=subpath)
+        return super(PackageURL, self).__new__(
+            PackageURL,
+            type=type,
+            namespace=namespace,
+            name=name,
+            version=version,
+            qualifiers=qualifiers,
+            subpath=subpath,
+        )
 
     def __str__(self, *args, **kwargs):
         return self.to_string()
@@ -306,9 +296,13 @@ class PackageURL(namedtuple('PackageURL', _components)):
         Return a purl string built from components.
         """
         type, namespace, name, version, qualifiers, subpath = normalize(# NOQA
-            self.type, self.namespace, self.name, self.version,
-            self.qualifiers, self.subpath,
-            encode=True
+            self.type,
+            self.namespace,
+            self.name,
+            self.version,
+            self.qualifiers,
+            self.subpath,
+            encode=True,
         )
 
         purl = ['pkg:', type, '/']
@@ -364,9 +358,7 @@ class PackageURL(namedtuple('PackageURL', _components)):
         if scheme or authority:
             msg = ('Invalid purl {} cannot contain a "user:pass@host:port" '
                    'URL Authority component: {}.')
-            raise ValueError(msg.format(
-                repr(purl), repr(authority)
-                                        ))
+            raise ValueError(msg.format(repr(purl), repr(authority)))
 
         path = path.lstrip('/')
         remainder, sep, version = path.rpartition('@')
@@ -392,8 +384,13 @@ class PackageURL(namedtuple('PackageURL', _components)):
                 'name component: {}'.format(repr(purl)))
 
         type, namespace, name, version, qualifiers, subpath = normalize(# NOQA
-            type, namespace, name, version, qualifiers, subpath,
-            encode=False
+            type,
+            namespace,
+            name,
+            version,
+            qualifiers,
+            subpath,
+            encode=False,
         )
 
         return PackageURL(type, namespace, name, version, qualifiers, subpath)
