@@ -24,25 +24,14 @@
 # Visit https://github.com/package-url/packageurl-python for support and
 # download.
 
-
-from __future__ import absolute_import
-from __future__ import print_function
-from __future__ import unicode_literals
-
 import os
 import re
-
-try:
-    from urlparse import urlparse  # Python 2
-    from urllib import unquote_plus
-except ImportError:
-    from urllib.parse import urlparse  # Python 3
-    from urllib.parse import unquote_plus
+from urllib.parse import unquote_plus
+from urllib.parse import urlparse
 
 from packageurl import PackageURL
-from packageurl.contrib.route import Router
 from packageurl.contrib.route import NoRouteAvailable
-
+from packageurl.contrib.route import Router
 
 """
 This module helps build a PackageURL from an arbitrary URL.
@@ -64,7 +53,9 @@ def url2purl(url):
         try:
             return purl_router.process(url)
         except NoRouteAvailable:
-            return
+            # If `url` does not fit in one of the existing routes,
+            # we attempt to create a generic PackageURL for `url`
+            return build_generic_purl(url)
 
 
 get_purl = url2purl
@@ -83,24 +74,54 @@ def purl_from_pattern(type_, pattern, url):
         return PackageURL(type_, **purl_data)
 
 
+def register_pattern(type_, pattern, router=purl_router):
+    """
+    Register a pattern with its type.
+    """
+    def endpoint(url):
+        return purl_from_pattern(type_, pattern, url)
+    router.append(pattern, endpoint)
+
+
 def get_path_segments(url):
     """
-    Return a list of path segments from a `url` string. This list may be empty. 
+    Return a list of path segments from a `url` string.
     """
     path = unquote_plus(urlparse(url).path)
     segments = [seg for seg in path.split("/") if seg]
-
-    if len(segments) <= 1:
-        segments = []
-
     return segments
 
 
+def build_generic_purl(uri):
+    """
+    Return a PackageURL from `uri`, if `uri` is a parsable URL, or None
+
+    `uri` is assumed to be a download URL, e.g. http://example.com/example.tar.gz
+    """
+    parsed_uri = urlparse(uri)
+    if parsed_uri.scheme and parsed_uri.netloc and parsed_uri.path:
+        # Get file name from `uri`
+        uri_path_segments = get_path_segments(uri)
+        if uri_path_segments:
+            file_name = uri_path_segments[-1]
+            return PackageURL(
+                type='generic',
+                name=file_name,
+                qualifiers={
+                    'download_url': uri
+                }
+            )
+
+
 @purl_router.route('https?://registry.npmjs.*/.*',
-                   'https?://registry.yarnpkg.com/.*')
+                   'https?://registry.yarnpkg.com/.*',
+                   'https?://(www\\.)?npmjs.*/package.*',
+                   'https?://(www\\.)?yarnpkg.com/package.*')
 def build_npm_purl(uri):
     # npm URLs are difficult to disambiguate with regex
-    if '/-/' in uri:
+    if '/package/' in uri:
+        return build_npm_web_purl(uri)
+    elif '/-/' in uri:
         return build_npm_download_purl(uri)
     else:
         return build_npm_api_purl(uri)
@@ -149,6 +170,42 @@ def build_npm_download_purl(uri):
     return PackageURL('npm', namespace, name, version)
 
 
+def build_npm_web_purl(uri):
+    path = unquote_plus(urlparse(uri).path)
+    if path.startswith('/package/'):
+        path = path[9:]
+
+    segments = [seg for seg in path.split('/') if seg]
+    len_segments = len(segments)
+    namespace = version = None
+
+    # @angular/cli/v/10.1.2
+    if len_segments == 4:
+        namespace = segments[0]
+        name = segments[1]
+        version = segments[3]
+
+    # express/v/4.17.1
+    elif len_segments == 3:
+        namespace = None
+        name = segments[0]
+        version = segments[2]
+
+    # @angular/cli
+    elif len_segments == 2:
+        namespace = segments[0]
+        name = segments[1]
+
+    # express
+    elif len_segments == 1 and len(segments) > 0 and segments[0][0] != '@':
+        name = segments[0]
+
+    else:
+        return
+
+    return PackageURL('npm', namespace, name, version)
+
+
 @purl_router.route('https?://repo1.maven.org/maven2/.*',
                    'https?://central.maven.org/maven2/.*',
                    'maven-index://repo1.maven.org/.*')
@@ -188,23 +245,30 @@ def build_maven_purl(uri):
     return PackageURL('maven', namespace, name, version, qualifiers)
 
 
-# https://rubygems.org/downloads/jwt-0.1.8.gem
-rubygems_pattern = (
-    r"^https?://rubygems.org/downloads/"
-    r"(?P<name>.+)-(?P<version>.+)"
-    r"(\.gem)$"
-)
-
-
-@purl_router.route(rubygems_pattern)
+@purl_router.route('https?://rubygems.org/downloads/.*')
 def build_rubygems_purl(uri):
+    # We use a more general route pattern instead of using `rubygems_pattern`
+    # below by itself because we want to capture all rubygems download URLs,
+    # even the ones that are not completly formed. This helps prevent url2purl
+    # from attempting to create a generic PackageURL from an invalid rubygems
+    # download URL.
+
+    # https://rubygems.org/downloads/jwt-0.1.8.gem
+    rubygems_pattern = (
+        r"^https?://rubygems.org/downloads/"
+        r"(?P<name>.+)-(?P<version>.+)"
+        r"(\.gem)$"
+    )
     return purl_from_pattern('rubygems', rubygems_pattern, uri)
 
 
 # https://pypi.python.org/packages/source/a/anyjson/anyjson-0.3.3.tar.gz
+# https://pypi.python.org/packages/2.6/t/threadpool/threadpool-1.2.7-py2.6.egg
+# https://pypi.python.org/packages/any/s/setuptools/setuptools-0.6c11-1.src.rpm
+# https://files.pythonhosted.org/packages/84/d8/451842a5496844bb5c7634b231a2e4caf0d867d2e25f09b840d3b07f3d4b/multi_key_dict-2.0.win32.exe
 pypi_pattern = (
-    r"(?P<name>.+)-(?P<version>.+)"
-    r"\.(zip|tar.gz|tar.bz2|.tgz)$"
+    r"(?P<name>(\w\.?)+(-\w+)*)-(?P<version>.+)"
+    r"\.(zip|tar.gz|tar.bz2|tgz|egg|rpm|exe)$"
 )
 
 # This pattern can be found in the following locations:
@@ -240,79 +304,80 @@ def build_pypi_purl(uri):
 
 # http://nuget.org/packages/EntityFramework/4.2.0.0
 # https://www.nuget.org/api/v2/package/Newtonsoft.Json/11.0.1
-nuget_pattern1 = (
+nuget_www_pattern = (
     r"^https?://.*nuget.org/(api/v2/)?packages?/"
     r"(?P<name>.+)/"
     r"(?P<version>.+)$"
 )
 
-
-@purl_router.route(nuget_pattern1)
-def build_nuget_purl1(uri):
-    return purl_from_pattern('nuget', nuget_pattern1, uri)
+register_pattern('nuget', nuget_www_pattern)
 
 
 # https://api.nuget.org/v3-flatcontainer/newtonsoft.json/10.0.1/newtonsoft.json.10.0.1.nupkg
-nuget_pattern2 = (
+nuget_api_pattern = (
     r"^https?://api.nuget.org/v3-flatcontainer/"
     r"(?P<name>.+)/"
     r"(?P<version>.+)/"
     r".*(nupkg)$"  # ends with "nupkg"
 )
 
-
-@purl_router.route(nuget_pattern2)
-def build_nuget_purl2(uri):
-    return purl_from_pattern('nuget', nuget_pattern2, uri)
+register_pattern('nuget', nuget_api_pattern)
 
 
-# http://master.dl.sourceforge.net/project/libpng/zlib/1.2.3/zlib-1.2.3.tar.bz2
-sourceforge_pattern = (
-    r"^https?://.*sourceforge.net/project/"
-    r"(?P<namespace>([^/]+))/"  # do not allow more "/" segments
-    r"(?P<name>.+)/"
-    r"(?P<version>[0-9\.]+)/"  # version restricted to digits and dots
-    r"(?P=name)-(?P=version).*"  # {name}-{version} repeated in the filename
-    r"[^/]$"  # not ending with "/"
-)
-
-
-@purl_router.route(sourceforge_pattern)
+@purl_router.route('https?://.*sourceforge.net/project/.*')
 def build_sourceforge_purl(uri):
-    return purl_from_pattern('sourceforge', sourceforge_pattern, uri)
+    # We use a more general route pattern instead of using `sourceforge_pattern`
+    # below by itself because we want to capture all sourceforge download URLs,
+    # even the ones that do not fit `sourceforge_pattern`. This helps prevent
+    # url2purl from attempting to create a generic PackageURL from a sourceforge
+    # URL that we can't handle.
+
+    # http://master.dl.sourceforge.net/project/libpng/zlib/1.2.3/zlib-1.2.3.tar.bz2
+    sourceforge_pattern = (
+        r"^https?://.*sourceforge.net/project/"
+        r"(?P<namespace>([^/]+))/"  # do not allow more "/" segments
+        r"(?P<name>.+)/"
+        r"(?P<version>[0-9\.]+)/"  # version restricted to digits and dots
+        r"(?P=name)-(?P=version).*"  # {name}-{version} repeated in the filename
+        r"[^/]$"  # not ending with "/"
+    )
+
+    sourceforge_purl = purl_from_pattern('sourceforge', sourceforge_pattern, uri)
+
+    if not sourceforge_purl:
+        # Get the project name from `uri` and use that as the Package name
+        # http://master.dl.sourceforge.net/project/aloyscore/aloyscore/0.1a1%2520stable/0.1a1_stable_AloysCore.zip
+        split_uri = uri.split('/project/') # http://master.dl.sourceforge.net, aloyscore/aloyscore/0.1a1%2520stable/0.1a1_stable_AloysCore.zip
+        if len(split_uri) >= 2:
+            remaining_uri_path = split_uri[1] # aloyscore/aloyscore/0.1a1%2520stable/0.1a1_stable_AloysCore.zip
+            remaining_uri_path_segments = remaining_uri_path.split('/') # aloyscore, aloyscore, 0.1a1%2520stable, 0.1a1_stable_AloysCore.zip
+            if remaining_uri_path_segments:
+                project_name = remaining_uri_path_segments[0] # aloyscore
+                sourceforge_purl = PackageURL(
+                    type='sourceforge',
+                    name=project_name,
+                    qualifiers={'download_url': uri}
+                )
+    return sourceforge_purl
 
 
+# https://crates.io/api/v1/crates/rand/0.7.2/download
 cargo_pattern = (
     r"^https?://crates.io/api/v1/crates/"
     r"(?P<name>.+)/(?P<version>.+)"
     r"(\/download)$"
 )
 
-
-@purl_router.route(cargo_pattern)
-def build_cargo_purl(url):
-    """
-    Return a PackageURL from a crates.io URL.
-    For example: https://crates.io/api/v1/crates/rand/0.7.2/download
-    """
-    return purl_from_pattern(type_='cargo', pattern=cargo_pattern, url=url)
+register_pattern('cargo', cargo_pattern)
 
 
+# https://raw.githubusercontent.com/volatilityfoundation/dwarf2json/master/LICENSE.txt
 github_raw_content_pattern = (
-    r"^https?://raw.githubusercontent.com/"
-    r"(?P<namespace>.+)/(?P<name>.+)/(?P<version>.+)/"
-    r"(?P<subpath>.+)$"
+    r"https?://raw.githubusercontent.com/(?P<namespace>[^/]+)/(?P<name>[^/]+)/"
+    r"(?P<version>[^/]+)/(?P<subpath>.*)$"
 )
 
-
-@purl_router.route(github_raw_content_pattern)
-def build_github_raw_content_purl(url):
-    """
-    Return a PackageURL object from GitHub Raw Content `url`.
-    For example:
-    https://raw.githubusercontent.com/volatilityfoundation/dwarf2json/master/LICENSE.txt
-    """
-    return purl_from_pattern(type_='github', pattern=github_raw_content_pattern, url=url)
+register_pattern('github', github_raw_content_pattern)
 
 
 @purl_router.route("https?://api.github\\.com/repos/.*")
@@ -321,7 +386,7 @@ def build_github_api_purl(url):
     Return a PackageURL object from GitHub API `url`.
     For example:
     https://api.github.com/repos/nexB/scancode-toolkit/commits/40593af0df6c8378d2b180324b97cb439fa11d66
-    https://api.github.com/repos/nexB/scancode-toolkit/ 
+    https://api.github.com/repos/nexB/scancode-toolkit/
     and returns a `PackageURL` object
     """
     segments = get_path_segments(url)
@@ -332,11 +397,11 @@ def build_github_api_purl(url):
     name = segments[2]
     version = None
 
-    #https://api.github.com/repos/nexB/scancode-toolkit/
+    # https://api.github.com/repos/nexB/scancode-toolkit/
     if len(segments) == 4 and segments[3] != 'commits':
         version = segments[3]
 
-    #https://api.github.com/repos/nexB/scancode-toolkit/commits/40593af0df6c8378d2b180324b97cb439fa11d66
+    # https://api.github.com/repos/nexB/scancode-toolkit/commits/40593af0df6c8378d2b180324b97cb439fa11d66
     if len(segments) == 5 and segments[3] == "commits":
         version = segments[4]
 
@@ -345,36 +410,36 @@ def build_github_api_purl(url):
     )
 
 
+# https://codeload.github.com/nexB/scancode-toolkit/tar.gz/v3.1.1
+# https://codeload.github.com/berngp/grails-rest/zip/release/0.7
 github_codeload_pattern = (
-    r"https?://codeload.github.com/"
-    r"(?P<namespace>.+)/(?P<name>.+)/(zip|tar.gz|tar.bz2|.tgz)/v?(?P<version>.+)$"
+    r"https?://codeload.github.com/(?P<namespace>.+)/(?P<name>.+)/"
+    r"(zip|tar.gz|tar.bz2|tgz)/(.*/)*"
+    r"v?(?P<version>.+)$"
 )
 
-
-@purl_router.route(github_codeload_pattern)
-def build_github_codeload_purl(url):
-    """
-    Return a PackageURL object from GitHub codeload `url`.
-    For example:
-    https://codeload.github.com/nexB/scancode-toolkit/tar.gz/v3.1.1
-    """
-    return purl_from_pattern(type_='github', pattern=github_codeload_pattern, url=url)
+register_pattern('github', github_codeload_pattern)
 
 
 @purl_router.route("https?://github\\.com/.*")
 def build_github_purl(url):
     """
     Return a PackageURL object from GitHub `url`.
-    For example:
-    https://github.com/package-url/packageurl-js/tree/master/test/data or
-    https://github.com/package-url/packageurl-js/tree/master or
-    https://github.com/package-url/packageurl-js or
-    https://github.com/nexB/scancode-toolkit/archive/v3.1.1.zip
     """
+
     # https://github.com/nexB/scancode-toolkit/archive/v3.1.1.zip
     archive_pattern = (
         r"https?://github.com/(?P<namespace>.+)/(?P<name>.+)"
-        r"/archive/v?(?P<version>.+).(zip|tar.gz|tar.bz2|.tgz)"
+        r"/archive/(.*/)*"
+        r"((?P=name)(-|_|@))?"
+        r"v?(?P<version>.+).(zip|tar.gz|tar.bz2|.tgz)"
+    )
+
+    # https://github.com/downloads/mozilla/rhino/rhino1_7R4.zip
+    download_pattern = (
+        r"https?://github.com/downloads/(?P<namespace>.+)/(?P<name>.+)/"
+        r"((?P=name)(-|@)?)?"
+        r"v?(?P<version>.+).(zip|tar.gz|tar.bz2|.tgz)"
     )
 
     # https://github.com/pypa/get-virtualenv/raw/20.0.31/public/virtualenv.pyz
@@ -383,13 +448,47 @@ def build_github_purl(url):
         r"/raw/v?(?P<version>[^/]+)/(?P<subpath>.*)$"
     )
 
-    for pattern in [archive_pattern, raw_pattern]:
+    # https://github.com/fanf2/unifdef/blob/master/unifdef.c
+    blob_pattern = (
+        r"https?://github.com/(?P<namespace>.+)/(?P<name>.+)"
+        r"/blob/(?P<version>[^/]+)/(?P<subpath>.*)$"
+    )
+
+    releases_download_pattern = (
+        r"https?://github.com/(?P<namespace>.+)/(?P<name>.+)"
+        r"/releases/download/v?(?P<version>[^/]+)/.*$"
+    )
+
+    # https://github.com/pombredanne/schematics.git
+    git_pattern = r"https?://github.com/(?P<namespace>.+)/(?P<name>.+).(git)"
+
+    patterns = (
+        archive_pattern,
+        raw_pattern,
+        blob_pattern,
+        releases_download_pattern,
+        download_pattern,
+        git_pattern,
+    )
+
+    for pattern in patterns:
         matches = re.search(pattern, url)
+        qualifiers = {}
         if matches:
-            return purl_from_pattern(type_='github', pattern=pattern, url=url)
+            if pattern != releases_download_pattern:
+                return purl_from_pattern(type_='github', pattern=pattern, url=url)
+            qualifiers['download_url'] = url
+            purl = purl_from_pattern(type_='github', pattern=pattern, url=url)
+            return PackageURL(
+                type=purl.type,
+                name=purl.name,
+                namespace=purl.namespace,
+                version=purl.version,
+                qualifiers=qualifiers
+            )
 
     segments = get_path_segments(url)
-    if not segments:
+    if not len(segments) >= 2:
         return
 
     namespace = segments[0]
@@ -423,14 +522,32 @@ def build_bitbucket_purl(url):
     For example:
     https://bitbucket.org/TG1999/first_repo/src/master or
     https://bitbucket.org/TG1999/first_repo/src or
-    https://bitbucket.org/TG1999/first_repo/src/master/new_folder 
+    https://bitbucket.org/TG1999/first_repo/src/master/new_folder
     """
+
     segments = get_path_segments(url)
 
-    if not segments:
+    if not len(segments) >= 2:
         return
     namespace = segments[0]
     name = segments[1]
+
+    bitbucket_download_pattern = (
+        r"https?://bitbucket.org/"
+        r"(?P<namespace>.+)/(?P<name>.+)/downloads/(?P<version>.+).(zip|tar.gz|tar.bz2|.tgz|exe|msi)"
+    )
+    matches = re.search(bitbucket_download_pattern, url)
+
+    qualifiers = {}
+    if matches:
+        qualifiers['download_url'] = url
+        return PackageURL(
+            type='bitbucket',
+            namespace=namespace,
+            name=name,
+            qualifiers=qualifiers
+        )
+
     version = None
     subpath = None
 
@@ -465,7 +582,7 @@ def build_gitlab_purl(url):
     """
     segments = get_path_segments(url)
 
-    if not segments:
+    if not len(segments) >= 2:
         return
     namespace = segments[0]
     name = segments[1]
@@ -491,19 +608,28 @@ def build_gitlab_purl(url):
     )
 
 
-hackage_pattern= (
+# https://hackage.haskell.org/package/a50-0.5/a50-0.5.tar.gz
+hackage_pattern = (
     r"^https?://hackage.haskell.org/package/"
     r"(?P<name>.+)-(?P<version>.+)/"
     r"(?P=name)-(?P=version).*"
     r"[^/]$"
 )
 
+register_pattern('hackage', hackage_pattern)
 
-@purl_router.route(hackage_pattern)
-def build_hackage_purl(url):
-    """
-    Return a PackageURL object from GitHub Raw Content `url`.
-    For example:
-    https://hackage.haskell.org/package/a50-0.5/a50-0.5.tar.gz
-    """
-    return purl_from_pattern(type_='hackage', pattern=hackage_pattern, url=url)
+
+@purl_router.route('https?://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/.*')
+def build_generic_google_code_archive_purl(uri):
+    # https://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/android-notifier/android-notifier-desktop-0.5.1-1.i386.rpm
+    _, remaining_uri = uri.split('https://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/') # android-notifier/android-notifier-desktop-0.5.1-1.i386.rpm
+    if remaining_uri:
+        split_remaining_uri = remaining_uri.split("/") # android-notifier, android-notifier-desktop-0.5.1-1.i386.rpm
+        if split_remaining_uri:
+            name = split_remaining_uri[0] # android-notifier
+            return PackageURL(
+                type='generic',
+                namespace='code.google.com',
+                name=name,
+                qualifiers={'download_url': uri}
+            )
